/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#include "fggl/debug/logging.hpp"
#include "fggl/debug/pragmas.hpp"

#include "fggl/gfx/ogl/common.hpp"
#include "fggl/display/glfw/window.hpp"

#include <fggl/gfx/camera.hpp>
#include <fggl/gfx/ogl/renderer.hpp>
#include <fggl/gfx/paint.hpp>
#include "fggl/data/model.hpp"
#include "fggl/gfx/ogl4/fallback.hpp"

#include <glm/ext/matrix_clip_space.hpp>
#include <glm/ext/matrix_transform.hpp>
#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>

#include <memory>

extern "C" {

/**
 * Convert an OpenGL source enum to a string.
 *
 * list of sources taken from table 20.1, GL Spec 4.5
 * @param source
 * @return string representing the source
 */
constexpr auto fggl_ogl_source(GLenum source) -> const char * {
	switch (source) {
	case GL_DEBUG_SOURCE_API: return "GL";
	case GL_DEBUG_SOURCE_SHADER_COMPILER: return "GLSL compiler";
	case GL_DEBUG_SOURCE_WINDOW_SYSTEM: return "Windowing System";
	case GL_DEBUG_SOURCE_THIRD_PARTY: return "Third Party";
	case GL_DEBUG_SOURCE_APPLICATION: return "Application";
	default:
	case GL_DEBUG_SOURCE_OTHER: return "Other";
	}
	assert(false);
	return "unknown";
}

/**
 * Convert an OpenGL type enum to a string.
 *
 * list of sources taken from table 20.2, GL Spec 4.5
 * @param type
 * @return string representing the type
 */
constexpr auto static fggl_ogl_type(GLenum type) -> const char * {
	switch (type) {
	case GL_DEBUG_TYPE_ERROR: return "Error";
		break;
	case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR: return "Deprecated Behaviour";
		break;
	case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR: return "Undefined Behaviour";
		break;
	case GL_DEBUG_TYPE_PERFORMANCE: return "Performance";
		break;
	case GL_DEBUG_TYPE_PORTABILITY: return "Portability";
		break;
	case GL_DEBUG_TYPE_MARKER: return "Marker";
		break;
	case GL_DEBUG_TYPE_PUSH_GROUP: return "Push Group";
		break;
	case GL_DEBUG_TYPE_POP_GROUP: return "Pop Group";
		break;
	default:
	case GL_DEBUG_TYPE_OTHER: return "Other";
		break;
	}
	assert(false);
	return "unknown";
}

PRAGMA_DIAGNOSTIC_PUSH
#pragma ide diagnostic ignored "bugprone-easily-swappable-parameters"

constexpr const char *OGL_LOG_FMT{"[GL] {}, {}: [{}]: {}"};

void static GLAPIENTRY fggl_ogl_log(GLenum source,
										  GLenum type,
										  unsigned int msgId,
										  GLenum severity,
										  GLsizei  /*length*/,
										  const char *message,
										  const void * /*userParam*/) {

	const auto *const sourceStr = fggl_ogl_source(source);
	const auto *const typeStr = fggl_ogl_type(type);

	// table 20.3, GL spec 4.5
	switch (severity) {
	case GL_DEBUG_SEVERITY_HIGH: fggl::debug::error(OGL_LOG_FMT, sourceStr, typeStr, msgId, message);
		break;
	default:
	case GL_DEBUG_SEVERITY_MEDIUM: fggl::debug::warning(OGL_LOG_FMT, sourceStr, typeStr, msgId, message);
		break;
	case GL_DEBUG_SEVERITY_LOW: fggl::debug::info(OGL_LOG_FMT, sourceStr, typeStr, msgId, message);
		break;
	case GL_DEBUG_SEVERITY_NOTIFICATION: fggl::debug::debug(OGL_LOG_FMT, sourceStr, typeStr, msgId, message);
		break;
	}
}

PRAGMA_DIAGNOSTIC_POP
}

/**
 * Future optimisations:
 *   recommended approach is to group stuff in to as few vao as possible - this
 * will do one vao per mesh, aka bad. Add support for instanced rendering
 * (particles) Look at packing vertex data in better ways (with profiling)
 *   Support shader specialisation (ie, dynamic/streamed data)
 *   Follow best recommendations for Vertex attributes (ie normals not using
 * floats, bytes for colour)
 *
 * Future features:
 *   Add support for models with weights (for animations)
 *   OpenGL ES for the FOSS thinkpad users who can't run anything even remotely
 * modern
 */
namespace fggl::gfx {

	using data::Mesh2D;
	using data::Vertex2D;

	OpenGL4Backend::OpenGL4Backend(data::Storage *storage, gui::FontLibrary *fonts, assets::AssetManager *assets, GlFunctionLoader loader)
		: fggl::gfx::Graphics(), m_canvasPipeline(nullptr), m_storage(storage) {

		// load OpenGL context, or fail.
		int version = gladLoadGLLoader(loader);
		if (version == 0) {
			debug::error("Failed to initialize OpenGL context\n");
			return;
		}

		// OpenGL debug Support
		GLint flags = 0;
		glGetIntegerv(GL_CONTEXT_FLAGS, &flags);
		if ((flags & GL_CONTEXT_FLAG_DEBUG_BIT) == GL_CONTEXT_FLAG_DEBUG_BIT) { // NOLINT(hicpp-signed-bitwise)
			debug::info("enabling OpenGL debug output");
			glEnable(GL_DEBUG_OUTPUT);
			glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
			glDebugMessageCallback(fggl_ogl_log, nullptr);
			glDebugMessageControl(GL_DONT_CARE, GL_DONT_CARE, GL_DONT_CARE, 0, nullptr, GL_TRUE);
		}

		// setup the viewport based on the window's framebuffer
		//auto fbSize = owner.frameSize();
		//glViewport(0, 0, fbSize.x, fbSize.y);

		// setup the shader cache
		m_cache = std::make_unique<ShaderCache>(m_storage);

		// setup 2D rendering system
		ShaderConfig shader2DConfig = ShaderConfig::named("canvas");
		m_canvasPipeline = m_cache->load(shader2DConfig);
		if (m_canvasPipeline == nullptr) {
			debug::error("failed to load shader2D - using fallback");
			m_canvasPipeline = m_cache->get(ogl4::FALLBACK_CANVAS_PIPELINE);
		}

		// FIXME this should not be hard-coded, it should be part of the scene loading process
		m_cache->load(ShaderConfig::named("phong"));
		m_cache->load(ShaderConfig::named("redbook/lighting"));
		m_cache->load(ShaderConfig::named("redbook/debug"));
		m_cache->load(ShaderConfig::named("normals", true));

		// rendering helpers
		m_canvasRenderer = std::make_unique<ogl4::CanvasRenderer>(fonts);
		m_modelRenderer = std::make_unique<ogl4::StaticModelRenderer>(m_cache.get(), assets);

		m_debugRenderer = std::make_unique<ogl4::DebugRenderer>(m_cache->getOrLoad(ShaderConfig::named("debug")));
		if (m_debugRenderer) {
			dd::initialize(m_debugRenderer.get());
		}
	};

	void OpenGL4Backend::clear() {
		glClearColor(0.0F, 0.0F, 0.0F, 0.0F);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	}

	void OpenGL4Backend::draw2D(const gfx::Paint &paint) {
		if (!m_canvasRenderer) {
			return;
		}

		m_canvasRenderer->render(*m_canvasPipeline, paint);
	}

	void OpenGL4Backend::drawScene(entity::EntityManager &world) {
		if (m_modelRenderer) {
			m_modelRenderer->render(world);
		}

		if (m_debugRenderer) {
			auto cameras = world.find<gfx::Camera>();
			if (cameras.empty()) {
				return;
			}

			auto cameraEnt = cameras.front();

			const auto &camTransform = world.get<math::Transform>(cameraEnt);
			const auto &camComp = world.get<gfx::Camera>(cameraEnt);

			const math::mat4 projectionMatrix =
				glm::perspective(camComp.fov, camComp.aspectRatio, camComp.nearPlane, camComp.farPlane);
			const math::mat4 viewMatrix = glm::lookAt(camTransform.origin(), camComp.target, camTransform.up());
			m_debugRenderer->mvpMatrix = projectionMatrix * viewMatrix;

			dd::flush();
		}
	}

	void OpenGL4Backend::resize(int width, int height) {
		glViewport(0, 0, width, height);
	}

}  // namespace fggl::gfx
