/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#include "fggl/gfx/ogl/types.hpp"
#include <cassert>

//
// special defines:
//
// FGGL_GL_I_BOUND will take responsibility away from fggl for ensuring buffers are bound before use.
// FGGL_GL_PARANOID will try to ensure that openGL state is managed correctly in method calls, but will be slower

namespace fggl::gfx::ogl {

	template<> const BuffAttrF attr_type<float>::attr = BuffAttrF::FLOAT;
	template<> const BuffAttrF attr_type<math::vec2>::attr = BuffAttrF::FLOAT;
	template<> const BuffAttrF attr_type<math::vec3>::attr = BuffAttrF::FLOAT;
	template<> const BuffAttrF attr_type<math::vec4>::attr = BuffAttrF::FLOAT;
	template<> const GLint attr_type<float>::size = 1;
	template<> const GLint attr_type<math::vec2>::size = 2;
	template<> const GLint attr_type<math::vec3>::size = 3;
	template<> const GLint attr_type<math::vec4>::size = 4;

	VertexArray::VertexArray() {
		glGenVertexArrays(1, &m_obj);
	}

	VertexArray::~VertexArray() {
		release();
	}

	VertexArray::VertexArray(VertexArray &&other) noexcept: m_obj(other.m_obj) {
		other.m_obj = 0;
	}

	VertexArray &VertexArray::operator=(VertexArray &&other) {
		if (this != &other) {
			release();
			std::swap(m_obj, other.m_obj);
		}
		return *this;
	}

	void VertexArray::release() {
		glDeleteVertexArrays(1, &m_obj);
		m_obj = 0;
	}

	void VertexArray::setAttribute(const ArrayBuffer &buff, GLuint idx, AttributeF &attr) {
		assert(0 <= idx && idx < GL_MAX_VERTEX_ATTRIBS);
		assert(1 <= attr.elmCount && attr.elmCount <= 4);
		assert(buff.isValid());

		#ifndef FGGL_GL_I_BOUND
		bind();
		GLuint boundVertexArray = 0;
		bind_buffer(&boundVertexArray, buff);
		#endif

		glEnableVertexAttribArray(idx);
		glVertexAttribPointer(idx, attr.elmCount, (GLenum) attr.attrType, GL_FALSE, attr.stride, (void *) attr.offset);

		#ifndef FGGL_GL_I_BOUND
		unbind_buffer(&boundVertexArray, buff);
		#endif
	}

	void VertexArray::setAttribute(const ArrayBuffer &buff, GLuint idx, AttributeI &attr, bool normalized) {
		assert(0 <= idx && idx < GL_MAX_VERTEX_ATTRIBS);
		assert(1 <= attr.elmCount && attr.elmCount <= 4);
		assert(buff.isValid());

		#ifndef FGGL_GL_I_BOUND
		GLuint boundVertexArray = 0;
		bind_buffer(&boundVertexArray, buff);
		#endif

		glVertexAttribPointer(idx,
							  attr.elmCount,
							  (GLenum) attr.attrType,
							  (GLboolean) normalized,
							  attr.stride,
							  (void *) attr.offset);

		#ifndef FGGL_GL_I_BOUND
		unbind_buffer(&boundVertexArray, buff);
		#endif
	}

	void VertexArray::setAttributeI(const ArrayBuffer &buff, GLuint idx, AttributeI &attr) {
		assert(0 <= idx && idx < GL_MAX_VERTEX_ATTRIBS);
		assert(1 <= attr.elmCount && attr.elmCount <= 4);
		assert(buff.isValid());

		#ifndef FGGL_GL_I_BOUND
		GLuint boundVertexArray = 0;
		bind_buffer(&boundVertexArray, buff);
		#endif

		glVertexAttribIPointer(idx, attr.elmCount, (GLenum) attr.attrType, attr.stride, (void *) attr.offset);

		#ifndef FGGL_GL_I_BOUND
		unbind_buffer(&boundVertexArray, buff);
		#endif
	}

	void VertexArray::drawElements(const ElementBuffer &buff, Primitive drawType, std::size_t size) {
		bind();

		#ifndef FGGL_I_BOUND
		GLuint boundElementArray = 0;
		bind_buffer(&boundElementArray, buff);
		#endif

		glDrawElements((GLenum) drawType, (GLsizei) size, GL_UNSIGNED_INT, nullptr);

		#ifndef FGGL_I_BOUND
		unbind_buffer(&boundElementArray, buff);
		#endif
	}

	void VertexArray::draw(Primitive drawType, int first, std::size_t count) {
		glDrawArrays((GLenum) drawType, first, count);
	}

} // namespace fggl::gfx::ogl