/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#include "fggl/gfx/ogl/common.hpp"
#include "fggl/gfx/ogl4/canvas.hpp"

#include "fggl/data/model.hpp"
#include "fggl/math/types.hpp"

#include <glm/gtc/type_ptr.hpp>
#include "fggl/math/triangulation.hpp"

#include "fggl/gui/fonts.hpp"

#define FGGL_OPENGL_CORRECTNESS

namespace fggl::gfx::ogl4 {

	static void make_box(gfx::Path2D &path, glm::vec2 topLeft, glm::vec2 bottomRight) {
		path.moveTo({topLeft.x, topLeft.y});
		path.pathTo({bottomRight.x, topLeft.y});
		path.pathTo({bottomRight.x, bottomRight.y});
		path.pathTo({topLeft.x, bottomRight.y});
		path.pathTo({topLeft.x, topLeft.y});
	}

	inline static void add_mesh_triangle(data::Mesh2D &mesh, const std::vector<data::Vertex2D> &verts) {
		assert(verts.size() == 3);
		for (const auto &vert : verts) {
			auto idx = mesh.add_vertex(vert);
			mesh.add_index(idx);
		}
	}

	static void convert_to_mesh(const gfx::Paint &paint, data::Mesh2D &mesh) {
		for (const auto &cmd : paint.cmds()) {
			auto path = cmd.path;

			std::vector<data::Vertex2D> verts;
			math::vec3 colour{1.0F, 1.0F, 1.0F};
			auto idx = 0;
			auto colourIdx = 0;

			for (auto &type : path.m_types) {
				if (type == PathType::PATH) {
					verts.push_back({path.m_points[idx++], colour});
				} else if (type == PathType::MOVE) {
					// polygon finished
					if (verts.size() < 3) {
						// empty, point, or line
						// TODO deal with whatever I'm meant to do with this...
					} else if (verts.size() == 3) {
						// triangle
						add_mesh_triangle(mesh, verts);
					} else {
						// polygon
						math::fan_triangulation(verts, mesh);
					}

					verts.clear();
					verts.push_back({path.m_points[idx++], colour});
				} else if (type == PathType::COLOUR) {
					colour = path.m_colours[colourIdx++];
				} else {
					// unsupported type
				}
			}

			if (!verts.empty() && verts.size() > 3) {
				math::fan_triangulation(verts, mesh);
			}
		}
	}

	CanvasRenderer::CanvasRenderer(fggl::gui::FontLibrary *fonts) :
		m_bounds({0.0F, 1920.F, 1080.0F, 0.0F}),
		m_fonts(fonts),
		m_fontTex(ogl::TextureType::Tex2D) {
		m_vao.bind();

		#ifdef FGGL_GL_I_BOUND
		// user will handle binding themselves usually, so attribute won't bind itself.
		// which means it's our problem right now...
		GLuint originalVertexList;
		ogl::bind_buffer( &originalVertexList, m_vertexList );
		#endif

		// define our attributes
		auto posAttr = ogl::attribute<data::Vertex2D, math::vec2>(offsetof(data::Vertex2D, position));
		auto colAttr = ogl::attribute<data::Vertex2D, math::vec3>(offsetof(data::Vertex2D, colour));
		auto texAttr = ogl::attribute<data::Vertex2D, math::vec2>(offsetof(data::Vertex2D, texPos));

		// bind the attributes to the vao
		m_vao.setAttribute(m_vertexList, 0, posAttr);
		m_vao.setAttribute(m_vertexList, 1, colAttr);
		m_vao.setAttribute(m_vertexList, 2, texAttr);

		#ifdef FGGL_GL_I_BOUND
		// cool, rebind whatever happened before, or not
		ogl::unbind_buffer( &originalVertexList, m_vertexList );
		#endif

		glBindVertexArray(0);
	}

	void CanvasRenderer::renderShapes(const gfx::Paint &paint, ogl::Shader& shader) {
		data::Mesh2D mesh;
		convert_to_mesh(paint, mesh);

		// nothing to render? give up
		if (mesh.indexList.empty()) {
			return;
		}

		// update data
		m_vao.bind();
		m_vertexList.replace(mesh.vertexList.size(), mesh.vertexList.data());
		m_indexList.replace(mesh.indexList.size(), mesh.indexList.data());

		// draw
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_CULL_FACE);

		auto projMat = glm::ortho(m_bounds.left, m_bounds.right, m_bounds.bottom, m_bounds.top);
		shader.use();
		shader.setUniformMtx(shader.uniform("projection"), projMat);

		// draw elements
		m_vao.drawElements(m_indexList, ogl::Primitive::TRIANGLE, mesh.indexList.size());

		// cleanup
		glUseProgram(0);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_CULL_FACE);
	}

	// slow version
	void CanvasRenderer::renderText(const Paint &paint, ogl::Shader& shader) {
		if (paint.textCmds().empty()) {
			return;
		}


		// get the expected font
		std::shared_ptr<gui::FontFace> face = m_fonts->getFont("LiberationSans-Regular.ttf");
		if (face == nullptr) {
			// we don't know about that font...
			return;
		}

		// setup the shader
		shader.use();
		auto projMat = glm::ortho(0.0f, 1920.0f, 1080.0f, 0.f);
		shader.setUniformMtx(shader.uniform("projection"), projMat);

		// bind the vbo we'll use for writing
		m_vao.bind();

		// setup the openGL state we expect for rendering
		glDisable(GL_DEPTH_TEST);
		glDisable(GL_CULL_FACE);

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		// for each text string, attempt to render it
		for (const auto &textCmd : paint.textCmds()) {
			const auto label = textCmd.text;
			math::vec2 penPos(textCmd.pos);

			for (auto letter : label) {
				ogl::Image img{};
				img.format = ogl::ImageFormat::R;
				img.type = ogl::PixelFormat::UNSIGNED_BYTE;

				glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
				face->texture(letter, img.size.x, img.size.y, &img.data);
				m_fontTex.setData(ogl::InternalImageFormat::Red, img);

				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

				m_fontTex.bind(1);
				shader.setUniformI(shader.uniform("tex"), 1);

				// this is why this is called the slow version, we render each quad as a single call
				auto &metrics = face->metrics(letter);
				const float xPos = penPos.x + metrics.bearing.x;
				const float yPos = (penPos.y - metrics.bearing.y);
				const float w = metrics.size.x;
				const float h = metrics.size.y;

				std::array<data::Vertex2D, 6> verts{{
														{{xPos, yPos + h}, textCmd.colour, {0.0F, 1.0F}},
														{{xPos, yPos}, textCmd.colour, {0.0F, 0.0F}},
														{{xPos + w, yPos}, textCmd.colour, {1.0F, 0.0F}},
														{{xPos, yPos + h}, textCmd.colour, {0.0F, 1.0F}},
														{{xPos + w, yPos}, textCmd.colour, {1.0F, 0.0F}},
														{{xPos + w, yPos + h}, textCmd.colour, {1.0F, 1.0F}},
													}};
				m_vertexList.replace(verts.size(), verts.data());

				m_vao.bind();
				m_vao.draw(ogl::Primitive::TRIANGLE, 0, verts.size());

				penPos.x += (metrics.advance >> 6);
			}
		}

		glDisable(GL_BLEND);
		glEnable(GL_DEPTH_TEST);
	}

	/*
	void renderTextWIP(const Paint& paint, GLuint shader) {

		if ( paint.textCmds().empty() ) {
			return;
		}

		auto fontFactory = util::ServiceLocator::instance().get<gui::FontLibrary>();
		std::shared_ptr<gui::FontFace> face = fontFactory->getFont("LiberationSans-Regular.ttf");
		if ( face == nullptr ){
			return;
		}

		Paint textPaint;
		for ( const auto& textCmd : paint.textCmds() ) {
			auto label = textCmd.text;

			gfx::Path2D textPath(textCmd.pos);
			math::vec2 penPos(textCmd.pos);

			// create a quad for this character
			for (auto letter : label) {
				auto& metrics = face->metrics(letter);

				std::cerr << metrics.size.x << std::endl;

				make_box(textPath, penPos, penPos + metrics.size);
				penPos.x += (metrics.advance >> 6 );
			}

			textPaint.fill(textPath);
		}

		data::Mesh2D mesh;
		convert_to_mesh(textPaint, mesh);

		// render the text mesh
		m_vao.bind();
		m_vertexList.replace(mesh.vertexList.size(), mesh.vertexList.data());
		m_indexList.replace(mesh.indexList.size(), mesh.indexList.data());

		// draw
		glUseProgram( shader );
		auto projMat = glm::ortho(0.0f, 1920.0f, 1080.0f, 0.f);
		glUniformMatrix4fv(glGetUniformLocation( shader, "projection"), 1, GL_FALSE,
						   glm::value_ptr(projMat));

		m_vao.drawElements(m_indexList, ogl::Primative::TRIANGLE, mesh.indexList.size());
		glUseProgram( 0 );
	}*/

	void CanvasRenderer::render(ogl::Shader& shader, const gfx::Paint &paint) {
		renderShapes(paint, shader);
		renderText(paint, shader);
	}

}
