/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 17/04/22.
//

#ifndef FGGL_GFX_OGL4_MODELS_HPP
#define FGGL_GFX_OGL4_MODELS_HPP

#include <string>
#include <unordered_map>

#include "fggl/gfx/ogl/shader.hpp"
#include "fggl/gfx/ogl/backend.hpp"
#include "fggl/gfx/ogl/types.hpp"

#include "fggl/data/model.hpp"
#include "fggl/assets/manager.hpp"

#define FGGL_ALLOW_DEFERRED_UPLOAD

namespace fggl::gfx::ogl4 {

	const std::size_t NO_RESTART_IDX = 0;

	struct StaticModelGPU {
		std::shared_ptr<ogl::VertexArray> vao;
		std::shared_ptr<ogl::ElementBuffer> elements;
		std::shared_ptr<ogl::ArrayBuffer> vertices;
		std::size_t elementCount;
		ogl::Primitive drawType = ogl::Primitive::TRIANGLE;
		std::size_t restartIndex = NO_RESTART_IDX;
	};

	struct StaticModelInstance {
		constexpr static auto name = "ogl::static::model";
		std::string modelName;
		StaticModelGPU* model = nullptr;
		std::shared_ptr<ogl::Shader> pipeline;
	};

	struct StaticModel {
		constexpr static auto name = "StaticModel";

		std::shared_ptr<ogl::Shader> pipeline;

		// element data
		std::shared_ptr<ogl::VertexArray> vao;
		std::shared_ptr<ogl::ElementBuffer> elements;
		std::shared_ptr<ogl::ArrayBuffer> vertexData;
		std::size_t elementCount;

		ogl::Primitive drawType;
		std::size_t restartIndex = NO_RESTART_IDX;
	};

	class StaticModelRenderer {
		public:
			inline StaticModelRenderer(gfx::ShaderCache *cache, assets::AssetManager *assets)
				: m_assets(assets), m_shaders(cache), m_phong(nullptr), m_vao(), m_vertexList(), m_indexList() {
				m_phong = cache->get("redbook/debug");
			}

			~StaticModelRenderer() = default;

			StaticModelRenderer(const StaticModelRenderer &other) = delete;
			StaticModelRenderer(StaticModelRenderer &&other) = delete;

			StaticModelRenderer &operator=(const StaticModelRenderer &other) = delete;
			StaticModelRenderer &operator=(StaticModelRenderer &&other) = delete;

			StaticModel* uploadMesh(assets::AssetGUID guid, const data::Mesh& mesh, bool allowCache=true);
			StaticModelGPU* uploadMesh2(const assets::AssetGUID& meshName, const data::Mesh& mesh);

			void render(entity::EntityManager &world) {
				#ifdef FGGL_ALLOW_DEFERRED_UPLOAD
					resolveModels(world);
				#endif
				renderModelsForward(world);
			}

			bool m_renderNormals = true;

		private:

			#ifdef FGGL_ALLOW_DEFERRED_UPLOAD
				/**
				 * Attach any missing rendering components to models.
				 */
				void resolveModels(entity::EntityManager &world);
			#endif

			/**
			 * Render all visible objects according to their render tokens.
			 */
			void renderModelsForward(const entity::EntityManager &world);

			assets::AssetManager *m_assets;
			gfx::ShaderCache *m_shaders;
			std::shared_ptr<ogl::Shader> m_phong;
			ogl::VertexArray m_vao;
			ogl::ArrayBuffer m_vertexList;
			ogl::ElementBuffer m_indexList;
	};

}

#endif //FGGL_GFX_OGL4_MODELS_HPP
